from django.db import models

from netkeiba.models.base import BaseModel


class Horse(BaseModel):
    UNKNOWN = 'UNKNOWN'
    MALE = 'MALE'
    FEMALE = 'FEMALE'
    CASTRATED = 'CASTRATED'
    SEX_CHOICES = (
        (MALE, 'Male'),
        (FEMALE, 'Female'),
        (CASTRATED, 'Castrated'),
        (UNKNOWN, 'Unknown')
    )

    key = models.CharField(max_length=255, unique=True)
    age = models.PositiveSmallIntegerField(null=True)
    sex = models.CharField(max_length=255, choices=SEX_CHOICES, default=UNKNOWN)
    name = models.CharField(max_length=255, null=True)

    class Meta:
        db_table = 'horses'



# Generated by Django 2.1.5 on 2019-02-07 12:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('netkeiba', '0002_add_models'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='race',
            name='grade',
        ),
    ]



import os
from datetime import datetime

import pytz
from django.core.management import BaseCommand
from scrapy.crawler import CrawlerRunner
from scrapy.settings import Settings
from twisted.internet import reactor

from netkeiba import settings
from netkeiba.argtype import date_string
from netkeiba.spiders.db_race import DBRaceSpider


class Command(BaseCommand):
    help = 'Start netkeiba db scrapy spider'

    def add_arguments(self, parser):
        parser.add_argument('--scrape-job-dirname', dest='scrapy_job_dirname',
                            help='Name to use as scrapy job directory. By default, this value is generated at runtime.')
        parser.add_argument('--min-date', dest='min_date', type=date_string,
                            help='Scrape all races that come on or after this date (fmt: YYYY-MM-DD)')
        parser.add_argument('--max-date', dest='max_date', type=date_string,
                            help='Scrape all races that come on or before this date (fmt: YYYY-MM-DD)')

    def handle(self, *args, **options):
        crawls_dir = os.path.join(settings.TMP_DIR, 'crawls')
        if not os.path.isdir(crawls_dir):
            os.makedirs(crawls_dir, exist_ok=True)

        piddir = os.path.join(settings.TMP_DIR, 'pids')
        if not os.path.isdir(piddir):
            os.makedirs(piddir, exist_ok=True)

        timestamp = datetime.now(tz=pytz.timezone(settings.TIME_ZONE)).strftime('%Y-%m-%dT%H%M%S')
        job_dirname = options.get('scrapy_job_dirname') if options.get('scrapy_job_dirname') else timestamp
        jobdir = os.path.join(crawls_dir, job_dirname)
        os.makedirs(jobdir)

        pidfile = os.path.join(piddir, f'{job_dirname}.pid')
        with open(pidfile, 'w') as f:
            f.write(str(os.getpid()) + os.linesep)

        custom_settings = {'JOBDIR': jobdir, }

        scrapy_settings = Settings()
        scrapy_settings.setmodule(settings, priority='project')

        spider = DBRaceSpider(min_date=options.get('min_date'), max_date=options.get('max_date'))
        runner = CrawlerRunner({**scrapy_settings, **custom_settings})
        d = runner.crawl(spider)
        d.addBoth(lambda _: reactor.stop())
        reactor.run()



import re
from datetime import datetime, date
from typing import List

import scrapy
from scrapy.link import Link
from scrapy.spiders import Rule
from scrapy.linkextractors import LinkExtractor

from netkeiba.items import WebPageItem


class DBRaceSpider(scrapy.Spider):
    name = 'db_race'
    allowed_domains = ['db.netkeiba.com']
    start_urls = ['https://db.netkeiba.com/?pid=race_top']

    rules = (
        Rule(LinkExtractor(allow=['/race/list/[0-9]+', 'pid=race_top&date=[0-9]+']),
             process_links='process_date_links'),
        Rule(LinkExtractor(allow='/race/[0-9]+', restrict_css='.race_list'), callback='parse_web_page_item',
             follow=True),
    )

    def __init__(self, min_date=None, max_date=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if min_date:
            self.min_date = datetime.strptime(min_date, '%Y-%m-%d').date()
        else:
            self.min_date = date(1970, 1, 1)

        if max_date:
            self.max_date = datetime.strptime(max_date, '%Y-%m-%d').date()
            start_url = self.start_urls[0]
            max_date_str = self.max_date.strftime('%Y%m%d')
            self.start_urls = ['&date='.join([start_url, max_date_str])]
        else:
            self.max_date = date.today()

        self.logger.info(f'<min_date: {self.min_date}, max_date: {self.max_date}, start_url: {self.start_urls[0]}>')

    def process_date_links(self, links: List[Link]):
        follow_links = []

        for link in links:
            date_string = re.search('[0-9]{8}', link.url).group()
            link_date = datetime.strptime(date_string, '%Y%m%d').date()

            is_date_in_range = self.min_date <= link_date <= self.max_date
            is_race_top_page = 'pid=race_top' in link.url
            is_month_above_limit = link_date.month >= self.min_date.month

            if is_date_in_range or (is_race_top_page and is_month_above_limit):
                follow_links.append(link)
            else:
                self.logger.info(f'skipping ({link.url}), ({self.min_date} <= {link_date} <= {self.max_date}) is False')

        return follow_links

    def parse(self, response):
        race_list_links = LinkExtractor(allow='/race/list/[0-9]+', restrict_css='.race_calendar') \
            .extract_links(response)

        for link in self.process_date_links(race_list_links):
            yield scrapy.Request(link.url, callback=self.parse_race_list)

        prev_links = LinkExtractor(allow='pid=race_top&date=[0-9]+', restrict_css='.race_calendar .rev') \
            .extract_links(response)

        valid_prev_links = self.process_date_links(prev_links)

        if len(valid_prev_links) > 1:
            yield scrapy.Request(valid_prev_links[-1].url, callback=self.parse)

    def parse_race_list(self, response):
        race_links = LinkExtractor(allow='/race/[0-9]+', restrict_css='.race_list').extract_links(response)

        for link in race_links:
            yield scrapy.Request(link.url, callback=self.scrape_web_page_item)

    def scrape_web_page_item(self, response):
        return WebPageItem.from_response(response)



from bs4 import BeautifulSoup


class Parser:
    def __init__(self, html):
        self._soup = BeautifulSoup(html, 'html.parser')
        self.data = None

    def parse(self):
        raise NotImplementedError

    def persist(self):
        raise NotImplementedError



from argparse import ArgumentTypeError
from datetime import datetime


def date_string(string):
    try:
        datetime.strptime(string, '%Y-%m-%d')
    except TypeError as e:
        raise ArgumentTypeError(e)
    return string



import logging
import os
from concurrent.futures import as_completed, ThreadPoolExecutor

from datetime import datetime

import pytz
from django.core.management import BaseCommand, CommandError

from netkeiba import settings
from netkeiba.models import WebPage

logger = logging.getLogger(__name__)


def import_page(queryset, i):
    page = queryset[i]
    parser = page.get_parser()
    try:
        parser.parse()
        parser.persist()
    except Exception as e:
        raise e
    return page.url


class Command(BaseCommand):
    help = 'Extract, clean and persist scraped netkeiba HTML'

    def add_arguments(self, parser):
        parser.add_argument('--scrapy-job-dirname', help='The name of the scrapy crawl jobdir')
        parser.add_argument('--offset', type=int, help='Start parsing web pages from {offset}', default=0)

    def _get_queryset(self, scrapy_job_dirname=None):
        if scrapy_job_dirname:
            requests_seen = os.path.join(settings.TMP_DIR, 'crawls', scrapy_job_dirname, 'requests.seen')

            if not os.path.exists(requests_seen):
                raise CommandError('jobdir/requests.seen does not exist')

            fingerprints = open(requests_seen).read().splitlines()
            queryset = WebPage.objects.filter(fingerprint__in=fingerprints)
        else:
            queryset = WebPage.objects.all()

        return queryset

    def handle(self, *args, **options):
        started_at = datetime.now(pytz.timezone(settings.TIME_ZONE))
        logger.info(f'START <{started_at}>')

        queryset = self._get_queryset(options.get('scrapy_job_dirname'))
        count = queryset.count()

        with ThreadPoolExecutor(max_workers=4) as executor:
            future_to_ix = {executor.submit(import_page, queryset, i): i for i in range(options['offset'], count)}
            for future in as_completed(future_to_ix):
                row = future_to_ix[future] + 1
                try:
                    url = future.result()
                except Exception as e:
                    logger.exception(e)
                else:
                    logger.info(f'({row}/{count}) <{url}>')

        stopped_at = datetime.now(pytz.timezone(settings.TIME_ZONE))
        duration = (stopped_at - started_at).seconds
        logger.info(f'STOP <{stopped_at}, duration: {duration} seconds>')



import os

from django.apps import apps
from django.conf import settings

DATABASES = getattr(settings, 'DATABASES')

APP_DIR = apps.get_app_config('netkeiba').path

TMP_DIR = getattr(settings, 'NETKEIBA_TMP_DIR', os.path.join(settings.BASE_DIR, 'tmp'))

TIME_ZONE = getattr(settings, 'NETKEIBA_TIME_ZONE', 'UTC')

BOT_NAME = 'netkeiba'
SPIDER_MODULES = ['netkeiba.spiders']
NEWSPIDER_MODULE = 'netkeiba.spiders'

ROBOTSTXT_OBEY = True

COOKIES_ENABLED = getattr(settings, 'NETKEIBA_COOKIES_ENABLED', False)

CRAWL_BFO = getattr(settings, 'NETKEIBA_CRAWL_BFO', False)

if CRAWL_BFO:
    DEPTH_PRIORITY = 1
    SCHEDULER_DISK_QUEUE = 'scrapy.squeues.PickleFifoDiskQueue'
    SCHEDULER_MEMORY_QUEUE = 'scrapy.squeues.FifoMemoryQueue'

AUTOTHROTTLE_ENABLED = True

CONCURRENT_REQUESTS_PER_DOMAIN = getattr(settings, 'NETKEIBA_CONCURRENT_REQUESTS_PER_DOMAIN', 6)

DOWNLOAD_DELAY = getattr(settings, 'NETKEIBA_DOWNLOAD_DELAY', 1.0)

RETRY_TIMES = getattr(settings, 'NETKEIBA_RETRY_TIMES', 4)

DOWNLOADER_MIDDLEWARES = {
    'netkeiba.middlewares.UserAgentMiddleware': 300,
}

ITEM_PIPELINES = {
    'netkeiba.pipelines.WebPagePipeline': 300,
}



from django.db import models

from netkeiba.models.base import BaseModel


class Owner(BaseModel):
    key = models.CharField(max_length=255, unique=True)
    name = models.CharField(max_length=255)

    class Meta:
        db_table = 'owners'



import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.1/howto/deployment/checklist/

# from django.core.management.utils import get_random_secret_key
# get_random_secret_key()
SECRET_KEY = os.getenv('SECRET_KEY')

ENVIRONMENT = os.getenv('ENVIRONMENT', 'development')

DEBUG = ENVIRONMENT == 'development'

ALLOWED_HOSTS = []

# Application definition

INSTALLED_APPS = ['netkeiba']

MIDDLEWARE = []

ROOT_URLCONF = 'config.urls'

TEMPLATES = []

# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    },
    'custom': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.getenv('CUSTOM_DB_PATH', ''),
    },
}

# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.1/howto/static-files/

STATIC_URL = '/static/'

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'default': {
            'format': '%(asctime)s %(levelname)-8s %(name)-15s %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
    },
    'filters': {
        'require_debug_true': {
            '()': 'django.utils.log.RequireDebugTrue',
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'filters': ['require_debug_true'],
            'level': 'DEBUG',
            'formatter': 'default',
        },
        'dlog': {
            'class': 'logging.FileHandler',
            'filters': ['require_debug_true'],
            'level': 'DEBUG',
            'formatter': 'default',
            'filename': 'debug.log',
        },
        'ilog': {
            'class': 'logging.FileHandler',
            'level': 'INFO',
            'formatter': 'default',
            'filename': 'info.log',
        },
    },
    'loggers': {
        '': {
            'handlers': ['console', 'dlog', 'ilog'],
            'level': 'DEBUG',
            'propagate': False
        }
    }
}



from .base import BaseModel
from .horse import Horse
from .jockey import Jockey
from .owner import Owner
from .race import Race
from .race_contender import RaceContender
from .trainer import Trainer
from .webpage import WebPage



import re
from datetime import datetime, time

import pytz
import scrapy
from bs4 import BeautifulSoup
from django.db import connection

from netkeiba.parsers.race import RACETRACKS, SURFACES, IMPOST_CATEGORIES, HORSE_SEX, WEATHER


class TABLE_COL:
    POST_POS = 0
    HORSE_NUM = 1
    HORSE = 3
    AGE_SEX = 4
    WEIGHT_CARRIED = 5
    JOCKEY = 6
    TRAINER = 7
    WEIGHT = 8
    ODDS = 9
    POPULARITY = 10


def _parse_racetrack(soup):
    return RACETRACKS.get(soup.select_one('.race_place a.active').text)


def _parse_datetime(soup):
    race_date = datetime.strptime(re.search('[0-9]{4}/[0-9]{2}/[0-9]{2}', soup.title.text).group(), '%Y/%m/%d').date()
    track_details = _parse_track_details(soup)
    time_str = re.search('[0-9]{2}:[0-9]{2}', track_details[-1]).group()
    hours, minutes = list(map(int, time_str.split(':')))
    race_time = time(hours, minutes)
    jst = pytz.timezone('Asia/Tokyo')
    dt = datetime.combine(race_date, race_time)
    return jst.localize(dt)


def _parse_track_details(soup):
    return soup.select_one('.racedata').select_one('p:nth-of-type(2)').text \
        .replace(u'\xa0', u'').replace(' ', '').split('/')


def _parse_course_type(soup):
    race_dist_type_data = soup.select_one('.racedata').select_one('p:nth-of-type(1)').text
    race_dist_type_search = re.search('(.)([0-9]+)m', race_dist_type_data)
    return SURFACES.get(race_dist_type_search.group(1))


def _parse_distance(soup):
    race_dist_type_data = soup.select_one('.racedata').select_one('p:nth-of-type(1)').text
    race_dist_type_search = re.search('(.)([0-9]+)m', race_dist_type_data)
    return int(race_dist_type_search.group(2))


def _parse_weather(soup):
    track_details = _parse_track_details(soup)
    weather = None
    for key, val in WEATHER.items():
        if key in track_details[0]:
            weather = val
    return weather


def _parse_track_condition(soup):
    track_details = _parse_track_details(soup)
    condition = None
    if '良' in track_details[1]:
        condition = 'good'
    elif '稍重' in track_details[1]:
        condition = 'slightly_heavy'
    elif '重' in track_details[1]:
        condition = 'heavy'
    elif '不良' in track_details[1]:
        condition = 'bad'
    return condition


def _parse_dirt_condition(soup):
    course_type = _parse_course_type(soup)
    if course_type == 'dirt':
        return _parse_track_condition(soup)
    return ''


def _parse_turf_condition(soup):
    course_type = _parse_course_type(soup)
    if course_type == 'turf':
        return _parse_track_condition(soup)
    return ''


def _parse_impost_category(soup):
    race_category_data = soup.select_one('.race_otherdata p:nth-of-type(2)').text
    for key, val in IMPOST_CATEGORIES.items():
        if key in race_category_data:
            return val
    return ''


def _parse_post_position(row):
    return row.select('td')[TABLE_COL.POST_POS].string


def _parse_horse_weight(row):
    weight_col = row.select('td')[TABLE_COL.WEIGHT].string
    weight_search = re.search('([0-9]+)\(([+-]?[0-9]+)\)', weight_col)
    return int(weight_search.group(1)) if weight_search else None


def _parse_horse_weight_diff(row):
    weight_col = row.select('td')[TABLE_COL.WEIGHT].string
    weight_search = re.search('([0-9]+)\(([+-]?[0-9]+)\)', weight_col)
    return int(weight_search.group(2)) if weight_search else None


def _parse_horse_url(row):
    return row.select('td')[TABLE_COL.HORSE].select_one('a').get('href')


def _parse_horse_key(row):
    h_url = _parse_horse_url(row)
    return re.search('/horse/([0-9]+)', h_url).group(1)


def _parse_horse_number(row):
    return int(row.select('td')[TABLE_COL.HORSE_NUM].string)


def _prefix_keys(obj: dict, prefix: str):
    acc = {}
    for key, value in obj.items():
        pref_key = ''.join([prefix, key])
        acc[pref_key] = value
    return acc


def _parse_horse_sex(row):
    horse_sex_symbol = re.search('[^0-9]', row.select('td')[TABLE_COL.AGE_SEX].text).group()
    return HORSE_SEX.get(horse_sex_symbol)


def _parse_weight_carried(row):
    return float(row.select('td')[TABLE_COL.WEIGHT_CARRIED].text)


def _parse_jockey_key(row):
    jockey_column = row.select('td')[TABLE_COL.JOCKEY]
    if jockey_column.select_one('a[href]'):
        j_url = jockey_column.select_one('a').get('href')
        j_key = re.search('/jockey/([0-9]+)', j_url).group(1)
    else:
        j_key = ''
    return j_key


def _parse_trainer_key(row):
    t_url = row.select('td')[TABLE_COL.TRAINER].select_one('a').get('href')
    return re.search('/trainer/([0-9]+)', t_url).group(1)


def _parse_odds(row):
    return float(row.select('td')[TABLE_COL.ODDS].text)


def _parse_popularity(row):
    return int(row.select('td')[TABLE_COL.POPULARITY].text)


def _parse_previous_order_of_finish(row, dt):
    h_key = _parse_horse_key(row)
    with connection.cursor() as cursor:
        previous_order_of_finish_query = '''
            SELECT c.order_of_finish
            FROM race_contenders c
                   LEFT JOIN races r ON c.race_id = r.id
            WHERE c.horse_id = (
              SELECT id
              FROM horses
              WHERE key = %s
              LIMIT 1
            )
              AND r.datetime < %s
            ORDER BY r.datetime DESC
            LIMIT 1
        '''
        result = cursor.execute(previous_order_of_finish_query, [h_key, dt]).fetchone()
        return result[0] if result else None


class RaceSpider(scrapy.Spider):
    name = 'race'

    allowed_domains = ['db.netkeiba.com']

    def __init__(self, race_url, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_urls = [race_url]
        self.logger.info(f'race_url set to {race_url}')

    def parse(self, response):
        soup = BeautifulSoup(response.text, 'html.parser')
        contenders = soup.select_one('.race_table_old').select('tr.bml1')
        contender_count = len(contenders)
        racetrack = _parse_racetrack(soup)
        datetime = _parse_datetime(soup)
        course_type = _parse_course_type(soup)
        distance = _parse_distance(soup)
        weather = _parse_weather(soup)
        dirt_condition = _parse_dirt_condition(soup)
        turf_condition = _parse_turf_condition(soup)
        impost_category = _parse_impost_category(soup)

        race_category_data = soup.select_one('.race_otherdata p:nth-of-type(2)').text
        is_non_winner_regional_horse_allowed = 1 if '(指定)' in race_category_data else 0
        is_winner_regional_horse_allowed = 1 if '特指' in race_category_data else 0
        is_regional_jockey_allowed = 1 if '[指定]' in race_category_data else 0
        is_foreign_horse_allowed = 1 if '混' in race_category_data else 0
        is_foreign_horse_and_trainer_allowed = 1 if '国際' in race_category_data else 0
        is_apprentice_jockey_allowed = 1 if '見習騎手' in race_category_data else 0
        is_female_only = 1 if '牝' in race_category_data else 0

        race = {
            'contender_count': contender_count,
            'racetrack': racetrack,
            'datetime': datetime,
            'course_type': course_type,
            'distance': distance,
            'weather': weather,
            'dirt_condition': dirt_condition,
            'turf_condition': turf_condition,
            'impost_category': impost_category,
            'is_non_winner_regional_horse_allowed': is_non_winner_regional_horse_allowed,
            'is_winner_regional_horse_allowed': is_winner_regional_horse_allowed,
            'is_regional_jockey_allowed': is_regional_jockey_allowed,
            'is_foreign_horse_allowed': is_foreign_horse_allowed,
            'is_foreign_horse_and_trainer_allowed': is_foreign_horse_and_trainer_allowed,
            'is_apprentice_jockey_allowed': is_apprentice_jockey_allowed,
            'is_female_only': is_female_only,
        }

        for row in contenders:
            c_post_position = _parse_post_position(row)
            h_url = _parse_horse_url(row)
            h_key = _parse_horse_key(row)
            h_sex = _parse_horse_sex(row)
            j_key = _parse_jockey_key(row)
            t_key = _parse_trainer_key(row)
            c_first_place_odds = _parse_odds(row)
            c_popularity = _parse_popularity(row)
            c_weight_carried = _parse_weight_carried(row)
            c_previous_order_of_finish = _parse_previous_order_of_finish(row, race['datetime'])
            c_horse_weight = _parse_horse_weight(row)
            c_horse_weight_diff = _parse_horse_weight_diff(row)
            c_horse_number = _parse_horse_number(row)

            meta = {
                'data': {
                    'c_post_position': c_post_position,
                    'h_key': h_key,
                    'h_sex': h_sex,
                    'j_key': j_key,
                    't_key': t_key,
                    'c_first_place_odds': c_first_place_odds,
                    'c_popularity': c_popularity,
                    'c_weight_carried': c_weight_carried,
                    'c_previous_order_of_finish': c_previous_order_of_finish,
                    'c_horse_weight': c_horse_weight,
                    'c_horse_weight_diff': c_horse_weight_diff,
                    'c_horse_number': c_horse_number,
                }
            }

            race_data = _prefix_keys(race, 'r_')
            meta['data'] = {**meta['data'], **race_data}
            yield {**meta['data'], **race_data}



import gzip
import logging
import os
import shutil
import sys
from datetime import datetime
from subprocess import PIPE, STDOUT, Popen, check_output

import pytz
from django.core.management import BaseCommand

from netkeiba import settings

logger = logging.getLogger(__name__)


def log_proc_output(out):
    with out:
        for line in out:
            logging.info(line)


class Command(BaseCommand):
    help = 'Compress sqlite db and upload to google cloud storage'

    def add_arguments(self, parser):
        parser.add_argument('-k', '--keep-backups', dest='keep_backups', type=int, default=3,
                            help='Number of backups to keep (old backups are deleted)')
        parser.add_argument('-l', '--local', dest='local', action='store_true', help='Output backup file and exit')

    def handle(self, *args, **options):
        started_at = datetime.now(tz=pytz.timezone(settings.TIME_ZONE))
        logger.info(f'Started backup command at {started_at}')

        db_path = settings.DATABASES['default']['NAME']
        db_gzip_path = f'{db_path}.gz'

        with open(db_path, 'rb') as f_in:
            with gzip.open(db_gzip_path, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

        if not options.get('local'):
            bucket_name = os.environ.get('GCLOUD_BUCKET')
            bucket_backup_dir = os.getenv('GCLOUD_BACKUP_DIR', 'data/db_backups')
            timestamp = datetime.now().strftime('%Y-%m-%dT%H%M%S')
            gcs_netkeiba_data_dir = os.path.join('gs://', bucket_name, bucket_backup_dir)
            gcs_backup_path = os.path.join(gcs_netkeiba_data_dir, timestamp, 'db.sqlite3.gz')

            cp_proc = Popen(['gsutil', 'cp', db_gzip_path, gcs_backup_path], stderr=STDOUT, stdout=PIPE)
            log_proc_output(cp_proc.stdout)

            backups = check_output(['gsutil', 'ls', gcs_netkeiba_data_dir], stderr=sys.stdout).splitlines()
            keep_backups = options.get('keep_backups')

            for backup in backups[:-keep_backups]:
                backup_path = backup.decode()
                rm_proc = Popen(['gsutil', 'rm', '-r', backup_path], stderr=STDOUT, stdout=PIPE)
                log_proc_output(rm_proc.stdout)

        stopped_at = datetime.now(tz=pytz.timezone(settings.TIME_ZONE))
        duration = (stopped_at - started_at).seconds
        logger.info(f'Finished backup command at {stopped_at} ({duration} seconds)')



import re

from django.db import models
from scrapy import Request
from scrapy.utils.request import request_fingerprint

from netkeiba.models.base import BaseModel
from netkeiba.parsers import Parser, RaceParser, NoopParser


class WebPageManager(models.Manager):
    url_regex = None
    parser_class = NoopParser

    def get_queryset(self):
        qs = super().get_queryset()
        if self.url_regex:
            qs = qs.filter(url__regex=self.url_regex)
        return qs

    def get_by_url(self, url):
        fingerprint = request_fingerprint(Request(url))
        return super().get_queryset().get(fingerprint=fingerprint)


class RaceWebPageManager(WebPageManager):
    """
    Ex. https://db.netkeiba.com/race/201807040211/
    """
    url_regex = '/race/[0-9]+/'
    parser_class = RaceParser


class WebPage(BaseModel):
    url = models.URLField()
    html = models.TextField()
    fingerprint = models.CharField(max_length=255, unique=True)
    crawled_at = models.DateTimeField()

    objects = WebPageManager()
    db_races = RaceWebPageManager()

    class Meta:
        db_table = 'webpages'

    @classmethod
    def _custom_managers(cls):
        return cls._meta.managers[1:]

    def get_parser(self) -> Parser:
        match = NoopParser
        for manager in self.__class__._custom_managers():
            if re.search(manager.url_regex, self.url):
                match = manager.parser_class
                break
        return match(self.html)



import os
import random

from . import settings


class UserAgentMiddleware(object):
    def __init__(self):
        filepath = os.path.join(settings.APP_DIR, 'user_agents.txt')

        with open(filepath) as f:
            self.user_agent_list = f.read().split('\n')

    def process_request(self, request, spider):
        user_agent = random.choice(self.user_agent_list)
        if user_agent:
            request.headers['User-Agent'] = user_agent

    def process_exception(self, request, exception, spider):
        user_agent = random.choice(self.user_agent_list)
        request.headers['User-Agent'] = user_agent



import pytz
from django.db import IntegrityError
from dateutil import parser as date_parser

from . import settings

from netkeiba.models import WebPage


class WebPagePipeline:
    def process_item(self, item, spider):
        tzinfo = pytz.timezone(settings.TIME_ZONE)
        crawled_at = date_parser.isoparse(item['crawled_at']).replace(tzinfo=tzinfo)
        defaults = {'html': item['html'], 'url': item['url'], 'crawled_at': crawled_at}

        try:
            WebPage.objects.create(fingerprint=item['fingerprint'], **defaults)
        except IntegrityError:
            spider.logger.debug(f"update <fingerprint: {item['fingerprint']}, url: {item['url']}>")
            WebPage.objects.filter(fingerprint=item['fingerprint']).update(**defaults)
        else:
            spider.logger.debug(f"insert <fingerprint: {item['fingerprint']}, url: {item['url']}>")

        return item



# Generated by Django 2.1.5 on 2019-02-07 00:52

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import re


class Migration(migrations.Migration):

    dependencies = [
        ('netkeiba', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Horse',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('key', models.CharField(max_length=255, unique=True)),
                ('age', models.PositiveSmallIntegerField(null=True)),
                ('sex', models.CharField(choices=[('MALE', 'Male'), ('FEMALE', 'Female'), ('CASTRATED', 'Castrated'), ('UNKNOWN', 'Unknown')], default='UNKNOWN', max_length=255)),
                ('name', models.CharField(max_length=255, null=True)),
            ],
            options={
                'db_table': 'horses',
            },
        ),
        migrations.CreateModel(
            name='Jockey',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('key', models.CharField(max_length=255, unique=True)),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'db_table': 'jockeys',
            },
        ),
        migrations.CreateModel(
            name='Owner',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('key', models.CharField(max_length=255, unique=True)),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'db_table': 'owners',
            },
        ),
        migrations.CreateModel(
            name='Race',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('key', models.CharField(max_length=255, unique=True)),
                ('racetrack', models.CharField(choices=[('SAPPORO', 'Sapporo'), ('HAKODATE', 'Hakodate'), ('FUMA', 'Fuma'), ('NIIGATA', 'Niigata'), ('TOKYO', 'Tokyo'), ('NAKAYAMA', 'Nakayama'), ('CHUKYO', 'Chukyo'), ('KYOTO', 'Kyoto'), ('HANSHIN', 'Hanshin'), ('OGURA', 'Ogura'), ('UNKNOWN', 'Unknown')], default='UNKNOWN', max_length=255)),
                ('impost_category', models.CharField(choices=[('HORSE_AGE', 'Horse age'), ('WEIGHT_FOR_AGE', 'Weight for age'), ('SET_WEIGHT', 'Set weight'), ('HANDICAP', 'Handicap'), ('UNKNOWN', 'Unknown')], default='UNKNOWN', max_length=255)),
                ('surface', models.CharField(choices=[('DIRT', 'Dirt'), ('TURF', 'Turf'), ('UNKNOWN', 'Unknown')], default='UNKNOWN', max_length=255)),
                ('course', models.CharField(choices=[('LEFT', 'Left'), ('RIGHT', 'Right'), ('STRAIGHT', 'Straight'), ('OBSTACLE', 'Obstacle'), ('UNKNOWN', 'Unknown')], default='UNKNOWN', max_length=255)),
                ('distance', models.PositiveSmallIntegerField()),
                ('number', models.PositiveSmallIntegerField()),
                ('race_class', models.CharField(choices=[('OPEN', 'Open'), ('U1600', 'Under 1600'), ('U1000', 'Under 1000'), ('U500', 'Under 500'), ('MAIDEN', 'Maiden'), ('UNRACED_MAIDEN', 'Unraced Maiden'), ('UNKNOWN', 'Unknown')], default='UNKNOWN', max_length=255)),
                ('grade', models.CharField(choices=[('G1', 'G1'), ('G2', 'G2'), ('G3', 'G3'), ('NA', 'Not applicable')], max_length=255)),
                ('datetime', models.DateTimeField()),
                ('weather', models.CharField(choices=[('CLOUDY', 'Cloudy'), ('RAINY', 'Rainy'), ('LIGHT_RAIN', 'Light rain'), ('SUNNY', 'Sunny'), ('SNOWY', 'Snowy'), ('UNKNOWN', 'Unknown')], default='UNKNOWN', max_length=255)),
                ('track_condition', models.CharField(choices=[('SLIGHTLY_HEAVY', 'Slightly heavy'), ('HEAVY', 'Heavy'), ('GOOD', 'Good'), ('BAD', 'Bad'), ('UNKNOWN', 'Unknown')], default='UNKNOWN', max_length=255)),
                ('is_outside_racetrack', models.BooleanField()),
                ('is_regional_maiden_race', models.BooleanField()),
                ('is_winner_regional_horse_race', models.BooleanField()),
                ('is_regional_jockey_race', models.BooleanField()),
                ('is_foreign_horse_race', models.BooleanField()),
                ('is_foreign_trainer_horse_race', models.BooleanField()),
                ('is_apprentice_jockey_race', models.BooleanField()),
                ('is_female_only_race', models.BooleanField()),
            ],
            options={
                'db_table': 'races',
                'get_latest_by': '-datetime',
            },
        ),
        migrations.CreateModel(
            name='RaceContender',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('order_of_finish', models.PositiveSmallIntegerField(null=True)),
                ('position_state', models.CharField(choices=[('OK', 'OK'), ('DISQUALIFIED', 'Disqualified'), ('CANCELLED', 'Cancelled'), ('REMOUNT', 'Remount'), ('NO_FINISH', 'Did not finish'), ('POSITION_LOWERED', 'Position lowered'), ('UNKNOWN', 'Unknown')], max_length=255)),
                ('post_position', models.PositiveSmallIntegerField()),
                ('horse_number', models.PositiveIntegerField()),
                ('weight_carried', models.FloatField()),
                ('finish_time', models.FloatField(null=True)),
                ('margin', models.CharField(choices=[('NOSE', 'Nose'), ('HEAD', 'Head'), ('NECK', 'Neck'), ('0.5', '0.5'), ('0.75', '0.75'), ('1.0', '1.0'), ('1.25', '1.25'), ('1.5', '1.5'), ('1.75', '1.75'), ('2.0', '2.0'), ('2.5', '2.5'), ('3.0', '3.0'), ('3.5', '3.5'), ('4.0', '4.0'), ('5.0', '5.0'), ('6.0', '6.0'), ('7.0', '7.0'), ('8.0', '8.0'), ('9.0', '9.0'), ('10.0', '10.0'), ('LARGE', 'Large'), ('UNKNOWN', 'Unknown')], max_length=255, null=True)),
                ('corner_pass_order', models.CharField(max_length=255, null=True, validators=[django.core.validators.RegexValidator(re.compile('^\\d+(?:,\\d+)*\\Z'), code='invalid', message='Enter only digits separated by commas.')])),
                ('final_stage_time', models.FloatField(null=True)),
                ('first_place_odds', models.FloatField(null=True)),
                ('popularity', models.PositiveSmallIntegerField(null=True)),
                ('horse_weight', models.FloatField(null=True)),
                ('horse_weight_diff', models.FloatField(null=True)),
                ('purse', models.FloatField(default=0.0)),
                ('horse', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='netkeiba.Horse')),
                ('jockey', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='netkeiba.Jockey')),
                ('owner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='netkeiba.Owner')),
                ('race', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='netkeiba.Race')),
            ],
            options={
                'db_table': 'race_contenders',
            },
        ),
        migrations.CreateModel(
            name='Trainer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('key', models.CharField(max_length=255, unique=True)),
                ('name', models.CharField(max_length=255)),
                ('stable', models.CharField(choices=[('EAST', 'East'), ('WEST', 'West'), ('REGIONAL', 'Regional'), ('OVERSEAS', 'Overseas'), ('UNKNOWN', 'Unknown')], default='UNKNOWN', max_length=255)),
            ],
            options={
                'db_table': 'trainers',
            },
        ),
        migrations.AddField(
            model_name='racecontender',
            name='trainer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='netkeiba.Trainer'),
        ),
        migrations.AlterUniqueTogether(
            name='race',
            unique_together={('racetrack', 'datetime')},
        ),
        migrations.AlterUniqueTogether(
            name='racecontender',
            unique_together={('race', 'horse', 'jockey', 'trainer', 'owner')},
        ),
    ]



from datetime import datetime

import pytz
import scrapy
from scrapy.utils.request import request_fingerprint

from . import settings


class WebPageItem(scrapy.Item):
    url = scrapy.Field()
    html = scrapy.Field()
    fingerprint = scrapy.Field()
    crawled_at = scrapy.Field()

    def __repr__(self):
        html = f"{self['html'][:40]}... (truncated)" if len(self['html']) > 55 else self['html']
        return repr({'url': self['url'], 'html': html})

    @classmethod
    def from_response(cls, response):
        fingerprint = request_fingerprint(response.request)
        crawled_at = datetime.now(pytz.timezone(settings.TIME_ZONE)).replace(microsecond=0).isoformat()
        return cls(url=response.url, html=response.text, fingerprint=fingerprint, crawled_at=crawled_at)



import logging
import re
from datetime import datetime, time, timedelta
from typing import Optional

import pytz
from bs4 import Comment
from django.db import IntegrityError

from netkeiba.models import Race, Horse, RaceContender, Trainer, Jockey, Owner
from netkeiba.parsers.parser import Parser

logger = logging.getLogger(__name__)

RACETRACKS = {
    '札幌': Race.SAPPORO,
    '函館': Race.HAKODATE,
    '福島': Race.FUMA,
    '新潟': Race.NIIGATA,
    '東京': Race.TOKYO,
    '中山': Race.NAKAYAMA,
    '中京': Race.CHUKYO,
    '京都': Race.KYOTO,
    '阪神': Race.HANSHIN,
    '小倉': Race.OGURA
}

WEATHER = {
    '曇': Race.CLOUDY,
    '晴': Race.SUNNY,
    '小雨': Race.LIGHT_RAIN,
    '雨': Race.RAINY,
    '雪': Race.SNOWY
}

IMPOST_CATEGORIES = {
    '馬齢': Race.HORSE_AGE,
    '定量': Race.WEIGHT_FOR_AGE,
    '別定': Race.SET_WEIGHT,
    'ハンデ': Race.HANDICAP
}

SURFACES = {
    '芝': Race.TURF,
    'ダ': Race.DIRT,
}

COURSE_TYPES = {
    '左': Race.LEFT,
    '右': Race.RIGHT,
    '直線': Race.STRAIGHT,
    '障': Race.OBSTACLE
}

HORSE_SEX = {
    '牝': Horse.FEMALE,
    '牡': Horse.MALE,
    'セ': Horse.CASTRATED
}

RACE_CLASSES = {
    'オープン': Race.OPEN,
    '1600万下': Race.U1600,
    '1000万下': Race.U1000,
    '500万下': Race.U500,
    '未勝利': Race.MAIDEN,
    '新馬': Race.UNRACED_MAIDEN,
}

POSITION_STATES = {
    '失': RaceContender.DISQUALIFIED,
    '取': RaceContender.CANCELLED,
    '除': RaceContender.CANCELLED,
    '再': RaceContender.REMOUNT,
    '中': RaceContender.NO_FINISH,
    '降': RaceContender.POSITION_LOWERED,
}

STABLES = {
    '東': Trainer.EAST,
    '西': Trainer.WEST,
    '地': Trainer.REGIONAL,
    '外': Trainer.OVERSEAS,
}

MARGINS = {
    'ハナ': RaceContender.NOSE,
    'クビ': RaceContender.NECK,
    'アタマ': RaceContender.HEAD,
    '1/2': RaceContender.BS_00__1_2,
    '3/4': RaceContender.BS_00__3_4,
    '1': RaceContender.BS_01__0_0,
    '1.1/4': RaceContender.BS_01__1_4,
    '1.1/2': RaceContender.BS_01__1_2,
    '1.3/4': RaceContender.BS_01__3_4,
    '2': RaceContender.BS_02__0_0,
    '2.1/2': RaceContender.BS_02__1_2,
    '3': RaceContender.BS_03__0_0,
    '3.1/2': RaceContender.BS_03__1_2,
    '4': RaceContender.BS_04__0_0,
    '5': RaceContender.BS_05__0_0,
    '6': RaceContender.BS_06__0_0,
    '7': RaceContender.BS_07__0_0,
    '8': RaceContender.BS_08__0_0,
    '9': RaceContender.BS_09__0_0,
    '10': RaceContender.BS_10__0_0,
    '大': RaceContender.HEAD,
}


# TODO: Parse other odds
class RaceParser(Parser):
    def parse(self):
        self.data = {
            'race': {
                'key': self._parse_key(),
                'racetrack': self._parse_racetrack(),
                'impost_category': self._parse_impost_category(),
                'surface': self._parse_surface(),
                'course': self._parse_course(),
                'distance': self._parse_distance(),
                'number': self._parse_number(),
                'race_class': self._parse_race_class(),
                'datetime': self._parse_datetime(),
                'weather': self._parse_weather(),
                'track_condition': self._parse_track_condition(),
                'is_outside_racetrack': self._parse_is_outside_racetrack(),
                'is_regional_maiden_race': self._parse_is_regional_maiden_race(),
                'is_winner_regional_horse_race': self._parse_is_winner_regional_horse_race(),
                'is_regional_jockey_race': self._parse_is_regional_jockey_race(),
                'is_foreign_horse_race': self._parse_is_foreign_horse_race(),
                'is_foreign_trainer_horse_race': self._parse_is_foreign_trainer_horse_race(),
                'is_apprentice_jockey_race': self._parse_is_apprentice_jockey_race(),
                'is_female_only_race': self._parse_is_female_only_race()
            }
        }

        contenders = []
        for i, record in enumerate(self._contender_rows):
            contenders.append({
                'horse': {
                    'key': self._parse_contender_horse_key(i),
                    'age': self._parse_contender_horse_age(i),
                    'sex': self._parse_contender_horse_sex(i),
                    'name': self._parse_contender_horse_name(i),
                },
                'jockey': {
                    'key': self._parse_contender_jockey_key(i),
                    'name': self._parse_contender_jockey_name(i),
                },
                'trainer': {
                    'key': self._parse_contender_trainer_key(i),
                    'name': self._parse_contender_trainer_name(i),
                    'stable': self._parse_contender_trainer_stable(i),
                },
                'owner': {
                    'key': self._parse_contender_owner_key(i),
                    'name': self._parse_contender_owner_name(i),
                },
                'order_of_finish': self._parse_contender_order_of_finish(i),
                'position_state': self._parse_contender_position_state(i),
                'post_position': self._parse_contender_post_position(i),
                'horse_number': self._parse_contender_horse_number(i),
                'weight_carried': self._parse_contender_weight_carried(i),
                'finish_time': self._parse_contender_finish_time(i),
                'margin': self._parse_contender_margin(i),
                'corner_pass_order': self._parse_contender_corner_pass_order(i),
                'final_stage_time': self._parse_contender_final_stage_time(i),
                'first_place_odds': self._parse_contender_first_place_odds(i),
                'popularity': self._parse_contender_popularity(i),
                'horse_weight': self._parse_contender_horse_weight(i),
                'horse_weight_diff': self._parse_contender_horse_weight_diff(i),
                'purse': self._parse_contender_purse(i),
            })
        self.data['contenders'] = contenders

    def persist(self):
        race_key = self.data['race']['key']

        try:
            Race.objects.create(**self.data['race'])
        except IntegrityError:
            Race.objects.filter(key=race_key).update(**self.data['race'])

        race = Race.objects.get(key=race_key)

        for contender in self.data['contenders']:
            horse, _ = Horse.objects.get_or_create(key=contender['horse']['key'], defaults={
                'age': contender['horse']['age'],
                'sex': contender['horse']['sex'],
                'name': contender['horse']['name']
            })

            jockey, _ = Jockey.objects.get_or_create(key=contender['jockey']['key'], defaults={
                'name': contender['jockey']['name']
            })

            trainer, _ = Trainer.objects.get_or_create(key=contender['trainer']['key'], defaults={
                'name': contender['trainer']['name'],
                'stable': contender['trainer']['stable']
            })

            owner, _ = Owner.objects.get_or_create(key=contender['owner']['key'], defaults={
                'name': contender['owner']['name'],
            })

            defaults = {
                'order_of_finish': contender['order_of_finish'],
                'position_state': contender['position_state'],
                'post_position': contender['post_position'],
                'horse_number': contender['horse_number'],
                'weight_carried': contender['weight_carried'],
                'finish_time': contender['finish_time'],
                'margin': contender['margin'],
                'corner_pass_order': contender['corner_pass_order'],
                'final_stage_time': contender['final_stage_time'],
                'first_place_odds': contender['first_place_odds'],
                'popularity': contender['popularity'],
                'horse_weight': contender['horse_weight'],
                'horse_weight_diff': contender['horse_weight_diff'],
                'purse': contender['purse'],
            }

            try:
                RaceContender.objects.create(race=race, horse=horse, jockey=jockey, trainer=trainer, owner=owner,
                                             **defaults)
            except IntegrityError:
                RaceContender.objects.filter(race=race, horse=horse, jockey=jockey, trainer=trainer,
                                             owner=owner).update(**defaults)

    @property
    def _track_details(self):
        return self._soup.select_one('.mainrace_data p span').string \
            .replace(u'\xa0', u'') \
            .replace(' ', '') \
            .split('/')

    @property
    def _subtitle(self):
        return self._soup.select_one('.mainrace_data .smalltxt').string \
            .replace(u'\xa0', u' ') \
            .replace('  ', ' ') \
            .split(' ')

    @property
    def _contender_rows(self):
        return self._soup.select('.race_table_01 tr')[1:]

    @property
    def _payoff_rows(self):
        return self._soup.select('.pay_block table tr')

    def _parse_key(self):
        race_url = self._soup.select_one('.race_num.fc .active').get('href')
        return re.search('/race/([0-9]+)', race_url).group(1)

    def _parse_racetrack(self):
        string = self._subtitle[1]
        match = Race.UNKNOWN
        for key, val in RACETRACKS.items():
            if re.search(key, string):
                match = val
                break
        return match

    def _parse_impost_category(self):
        string = self._subtitle[-1]
        match = Race.UNKNOWN
        for key, val in IMPOST_CATEGORIES.items():
            if key in string:
                match = val
                break
        return match

    def _parse_surface(self):
        string = self._track_details[0]
        match = Race.UNKNOWN
        for key, value in SURFACES.items():
            if key in string:
                match = value
                break
        return match

    def _parse_course(self):
        string = self._track_details[0]
        match = Race.UNKNOWN
        for key, value in COURSE_TYPES.items():
            if key in string:
                match = value
                break
        return match

    def _parse_is_outside_racetrack(self):
        return '外' in self._track_details[0]

    def _parse_distance(self):
        string = self._track_details[0]
        match = re.search('([0-9]+)m', string).group(1)
        return int(match)

    def _parse_number(self):
        string = self._soup.select_one('.mainrace_data .racedata dt').string.strip()
        return int(string.split(' ')[0])

    def _parse_race_class(self):
        header = self._soup.select_one('.mainrace_data h1')
        for node in header.children:
            if isinstance(node, Comment):
                node.extract()
        match = None
        for grade in [Race.G1, Race.G2, Race.G3]:
            if grade in header.text:
                match = grade
                break
        if match is None:
            race_terms = self._subtitle[-2]
            for key, val in RACE_CLASSES.items():
                if key in race_terms:
                    match = val
                    break
        return match if match else Race.UNKNOWN

    def _parse_datetime(self):
        race_date = self._parse_date()
        race_time = self._parse_time()
        jst = pytz.timezone('Asia/Tokyo')
        dt = datetime.combine(race_date, race_time)
        return jst.localize(dt)

    def _parse_date(self):
        return datetime.strptime(self._subtitle[0], '%Y年%m月%d日').date()

    def _parse_time(self):
        time_str = re.search('[0-9]{2}:[0-9]{2}', self._track_details[-1]).group()
        hours, minutes = list(map(int, time_str.split(':')))
        return time(hours, minutes)

    def _parse_weather(self):
        string = self._track_details[1]
        match = Race.UNKNOWN
        for key, val in WEATHER.items():
            if key in string:
                match = val
                break
        return match

    def _parse_track_condition(self):
        string = self._track_details[2]
        match = Race.UNKNOWN
        if '不良' in string:
            match = Race.BAD
        elif '良' in string:
            match = Race.GOOD
        elif '稍重' in string:
            match = Race.SLIGHTLY_HEAVY
        elif '重' in string:
            match = Race.HEAVY
        return match

    def _parse_is_regional_maiden_race(self):
        return '(指定)' in self._subtitle[-1]

    def _parse_is_winner_regional_horse_race(self):
        return '特指' in self._subtitle[-1]

    def _parse_is_regional_jockey_race(self):
        return '[指定]' in self._subtitle[-1]

    def _parse_is_foreign_horse_race(self):
        return '混' in self._subtitle[-1]

    def _parse_is_foreign_trainer_horse_race(self):
        return '国際' in self._subtitle[-1]

    def _parse_is_apprentice_jockey_race(self):
        return '見習騎手' in self._subtitle[-1]

    def _parse_is_female_only_race(self):
        return '牝' in self._subtitle[-1]

    def _parse_contender_horse_key(self, i):
        horse_url = self._contender_rows[i].select('td')[3].select_one('a').get('href')
        return re.search('/horse/([0-9]+)', horse_url).group(1)

    def _parse_contender_horse_age(self, i):
        string = self._contender_rows[i].select('td')[4].string
        return int(re.search('[0-9]+', string).group())

    def _parse_contender_horse_sex(self, i):
        string = self._contender_rows[i].select('td')[4].string
        match = Race.UNKNOWN
        for key, value in HORSE_SEX.items():
            if key in string:
                match = value
                break
        return match

    def _parse_contender_horse_name(self, i):
        string = self._contender_rows[i].select('td')[3].select_one('a').string
        return string if string else ''

    def _parse_contender_jockey_key(self, i):
        jockey_url = self._contender_rows[i].select('td')[6].select_one('a').get('href')
        return re.search('/jockey/([a-z0-9]+)', jockey_url).group(1)

    def _parse_contender_jockey_name(self, i):
        string = self._contender_rows[i].select('td')[6].select_one('a').string
        return string if string else ''

    def _parse_contender_trainer_key(self, i):
        trainer_url = self._contender_rows[i].select('td')[-3].select_one('a').get('href')
        return re.search('/trainer/([a-z0-9]+)', trainer_url).group(1)

    def _parse_contender_trainer_stable(self, i):
        string = self._contender_rows[i].select('td')[-3].text
        stable_char = re.search('\[(東|西|地|外)\]', string).group(1)
        match = Trainer.UNKNOWN
        for key, value in STABLES.items():
            if stable_char == key:
                match = key
                break
        return match

    def _parse_contender_trainer_name(self, i):
        string = self._contender_rows[i].select('td')[-3].select_one('a').string
        return string if string else ''

    def _parse_contender_owner_key(self, i):
        owner_link = self._contender_rows[i].select('td')[-2].select_one('a')
        if owner_link is None:
            raise ValueError('Owner does not have link. This is probably a really old race!')
        owner_url = owner_link.get('href')
        return re.search('/owner/([a-z0-9]+)', owner_url).group(1)

    def _parse_contender_owner_name(self, i):
        string = self._contender_rows[i].select('td')[-2].text.strip()
        return string if string else ''

    def _parse_contender_order_of_finish(self, i):
        string = self._contender_rows[i].select('td')[0].string
        match = re.search('[0-9]+', string)
        return match.group() if match else None

    def _parse_contender_position_state(self, i):
        string = self._contender_rows[i].select('td')[0].string
        match = RaceContender.OK
        for key, value in POSITION_STATES.items():
            if key in string:
                match = value
                break
        return match

    def _parse_contender_post_position(self, i):
        string = self._contender_rows[i].select('td')[1].string
        return int(string)

    def _parse_contender_weight_carried(self, i):
        string = self._contender_rows[i].select('td')[5].string
        return float(string)

    def _parse_contender_horse_number(self, i):
        string = self._contender_rows[i].select('td')[2].string
        return int(string)

    def _parse_contender_finish_time(self, i) -> Optional[float]:
        string = self._contender_rows[i].select('td')[7].string
        if string:
            dt = datetime.strptime(string, '%M:%S.%f')
            td = timedelta(minutes=dt.minute, seconds=dt.second, microseconds=dt.microsecond)
            return td.total_seconds()
        return None

    def _parse_contender_margin(self, i):
        string = self._contender_rows[i].select('td')[8].string
        match = RaceContender.UNKNOWN
        if string:
            for key, value in MARGINS.items():
                if string == key:
                    match = value
                    break
        return match

    def _parse_contender_corner_pass_order(self, i):
        string = self._contender_rows[i].select('td')[10].string
        if string:
            return string.replace('-', ',')
        return None

    def _parse_contender_final_stage_time(self, i):
        string = self._contender_rows[i].select('td')[11].string
        return float(string) if string else None

    def _parse_contender_first_place_odds(self, i):
        string = self._contender_rows[i].select('td')[12].string
        return None if string == '---' else float(string.replace(',', ''))

    def _parse_contender_popularity(self, i):
        string = self._contender_rows[i].select('td')[13].string
        return int(string) if string else None

    def _parse_contender_horse_weight(self, i) -> Optional[int]:
        string = self._contender_rows[i].select('td')[14].string
        match = re.search('([0-9]+)\(([+-]?[0-9]+)\)', string)
        return int(match.group(1)) if match else None

    def _parse_contender_horse_weight_diff(self, i) -> Optional[int]:
        string = self._contender_rows[i].select('td')[14].string
        match = re.search('([0-9]+)\(([+-]?[0-9]+)\)', string)
        return int(match.group(2)) if match else None

    def _parse_contender_purse(self, i):
        string = self._contender_rows[i].select('td')[-1].string
        return float(string.replace(',', '')) if string else 0.



from django.core.validators import validate_comma_separated_integer_list
from django.db import models

from netkeiba.models import BaseModel


class RaceContender(BaseModel):
    UNKNOWN = 'UNKNOWN'

    OK = 'OK'
    DISQUALIFIED = 'DISQUALIFIED'  # 失
    CANCELLED = 'CANCELLED'  # 取|除
    REMOUNT = 'REMOUNT'  # 再
    NO_FINISH = 'NO_FINISH'  # 中
    POSITION_LOWERED = 'POSITION_LOWERED'  # 降
    POSITION_STATE_CHOICES = (
        (OK, 'OK'),
        (DISQUALIFIED, 'Disqualified'),
        (CANCELLED, 'Cancelled'),
        (REMOUNT, 'Remount'),
        (NO_FINISH, 'Did not finish'),
        (POSITION_LOWERED, 'Position lowered'),
        (UNKNOWN, 'Unknown')
    )

    NOSE = 'NOSE'
    HEAD = 'HEAD'
    NECK = 'NECK'
    BS_00__1_2 = '0.5'
    BS_00__3_4 = '0.75'
    BS_01__0_0 = '1.0'
    BS_01__1_4 = '1.25'
    BS_01__1_2 = '1.5'
    BS_01__3_4 = '1.75'
    BS_02__0_0 = '2.0'
    BS_02__1_2 = '2.5'
    BS_03__0_0 = '3.0'
    BS_03__1_2 = '3.5'
    BS_04__0_0 = '4.0'
    BS_05__0_0 = '5.0'
    BS_06__0_0 = '6.0'
    BS_07__0_0 = '7.0'
    BS_08__0_0 = '8.0'
    BS_09__0_0 = '9.0'
    BS_10__0_0 = '10.0'
    LARGE = 'LARGE'
    MARGIN_CHOICES = (
        (NOSE, 'Nose'),
        (HEAD, 'Head'),
        (NECK, 'Neck'),
        (BS_00__1_2, '0.5'),
        (BS_00__3_4, '0.75'),
        (BS_01__0_0, '1.0'),
        (BS_01__1_4, '1.25'),
        (BS_01__1_2, '1.5'),
        (BS_01__3_4, '1.75'),
        (BS_02__0_0, '2.0'),
        (BS_02__1_2, '2.5'),
        (BS_03__0_0, '3.0'),
        (BS_03__1_2, '3.5'),
        (BS_04__0_0, '4.0'),
        (BS_05__0_0, '5.0'),
        (BS_06__0_0, '6.0'),
        (BS_07__0_0, '7.0'),
        (BS_08__0_0, '8.0'),
        (BS_09__0_0, '9.0'),
        (BS_10__0_0, '10.0'),
        (LARGE, 'Large'),
        (UNKNOWN, 'Unknown')
    )

    race = models.ForeignKey('netkeiba.Race', on_delete=models.CASCADE)
    horse = models.ForeignKey('netkeiba.Horse', on_delete=models.CASCADE)
    jockey = models.ForeignKey('netkeiba.Jockey', on_delete=models.CASCADE)
    trainer = models.ForeignKey('netkeiba.Trainer', on_delete=models.CASCADE)
    owner = models.ForeignKey('netkeiba.Owner', on_delete=models.CASCADE)
    order_of_finish = models.PositiveSmallIntegerField(null=True)
    position_state = models.CharField(max_length=255, choices=POSITION_STATE_CHOICES)
    post_position = models.PositiveSmallIntegerField()
    horse_number = models.PositiveIntegerField()
    weight_carried = models.FloatField()
    finish_time = models.FloatField(null=True)
    margin = models.CharField(max_length=255, choices=MARGIN_CHOICES, null=True)
    corner_pass_order = models.CharField(max_length=255, validators=[validate_comma_separated_integer_list], null=True)
    final_stage_time = models.FloatField(null=True)
    first_place_odds = models.FloatField(null=True)
    popularity = models.PositiveSmallIntegerField(null=True)
    horse_weight = models.FloatField(null=True)
    horse_weight_diff = models.FloatField(null=True)
    purse = models.FloatField(default=0.)

    class Meta:
        db_table = 'race_contenders'
        unique_together = ('race', 'horse', 'jockey', 'trainer', 'owner')



import logging
from datetime import datetime
from subprocess import call

import pytz
from django.conf import settings
from django.core.management import BaseCommand, call_command

from netkeiba.argtype import date_string

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Execute the netkeiba training pipeline (scrape, import, backup)'

    def add_arguments(self, parser):
        parser.add_argument('--min-date', dest='min_date', type=date_string,
                            help='Scrape all races that come on or after this date (fmt: YYYY-MM-DD)')
        parser.add_argument('--max-date', dest='max_date', type=date_string,
                            help='Scrape all races that come on or before this date (fmt: YYYY-MM-DD)')
        parser.add_argument('--shutdown', action='store_true', default=False,
                            help='Execute shutdown command after pipeline completion')
        parser.add_argument('--backup', action='store_true', default=False,
                            help='Make gzip copy of sqlite database after running scrape and import commands')

    def handle(self, *args, **options):
        started_at = datetime.now(tz=pytz.timezone(settings.TIME_ZONE))
        logger.info(f'Started pipeline command at {started_at}')

        scrapy_job_dirname = started_at.strftime('%Y-%m-%dT%H%M%S')
        logger.debug(f'Scrapy job dirname set to {scrapy_job_dirname}')

        call_command('scrape', scrapy_job_dirname=scrapy_job_dirname, min_date=options.get('min_date'),
                     max_date=options.get('max_date'))
        call_command('import', scrapy_job_dirname=scrapy_job_dirname)

        if options.get('backup'):
            call_command('backup')

        stopped_at = datetime.now(tz=pytz.timezone(settings.TIME_ZONE))
        duration = (stopped_at - started_at).seconds
        logger.info(f'Finished pipeline command at {stopped_at} ({duration} seconds)')

        if options.get('shutdown'):
            logger.info('Received shutdown option. Shutting down now.')
            call('sudo shutdown -h now', shell=True)



from django.db import models

from netkeiba.models.base import BaseModel


class Race(BaseModel):
    UNKNOWN = 'UNKNOWN'

    SAPPORO = 'SAPPORO'
    HAKODATE = 'HAKODATE'
    FUMA = 'FUMA'
    NIIGATA = 'NIIGATA'
    TOKYO = 'TOKYO'
    NAKAYAMA = 'NAKAYAMA'
    CHUKYO = 'CHUKYO'
    KYOTO = 'KYOTO'
    HANSHIN = 'HANSHIN'
    OGURA = 'OGURA'
    RACETRACK_CHOICES = (
        (SAPPORO, 'Sapporo'),
        (HAKODATE, 'Hakodate'),
        (FUMA, 'Fuma'),
        (NIIGATA, 'Niigata'),
        (TOKYO, 'Tokyo'),
        (NAKAYAMA, 'Nakayama'),
        (CHUKYO, 'Chukyo'),
        (KYOTO, 'Kyoto'),
        (HANSHIN, 'Hanshin'),
        (OGURA, 'Ogura'),
        (UNKNOWN, 'Unknown'),
    )

    HORSE_AGE = 'HORSE_AGE'
    WEIGHT_FOR_AGE = 'WEIGHT_FOR_AGE'
    SET_WEIGHT = 'SET_WEIGHT'
    HANDICAP = 'HANDICAP'
    IMPOST_CATEGORY_CHOICES = (
        # [馬齢] 同一年齢の馬だけのレース
        (HORSE_AGE, 'Horse age'),
        # [定量] 別定のなかでも、馬の年齢と性別を基準に定められているレース
        (WEIGHT_FOR_AGE, 'Weight for age'),
        # [別定] そのレース毎に負担重量を決定する基準が設けられているレース
        (SET_WEIGHT, 'Set weight'),
        # [ハンデ] 出走予定馬の実績や最近の状態などを考慮し
        # 各出走馬に勝つチャンスを与えるよう決められた重量を負担させるレース
        (HANDICAP, 'Handicap'),
        (UNKNOWN, 'Unknown'),
    )

    DIRT = 'DIRT'
    TURF = 'TURF'
    SURFACE_CHOICES = (
        (DIRT, 'Dirt'),
        (TURF, 'Turf'),
        (UNKNOWN, 'Unknown'),
    )

    G1 = 'G1'
    G2 = 'G2'
    G3 = 'G3'
    OPEN = 'OPEN'
    U1600 = 'U1600'
    U1000 = 'U1000'
    U500 = 'U500'
    MAIDEN = 'MAIDEN'
    UNRACED_MAIDEN = 'UNRACED_MAIDEN'
    RACE_CLASS_CHOICES = (
        (G1, 'G1'),
        (G2, 'G2'),
        (G3, 'G3'),
        (OPEN, 'Open'),
        (U1600, 'Under 1600'),
        (U1000, 'Under 1000'),
        (U500, 'Under 500'),
        (MAIDEN, 'Maiden'),
        (UNRACED_MAIDEN, 'Unraced Maiden'),
        (UNKNOWN, 'Unknown'),
    )

    CLOUDY = 'CLOUDY'
    RAINY = 'RAINY'
    LIGHT_RAIN = 'LIGHT_RAIN'
    SUNNY = 'SUNNY'
    SNOWY = 'SNOWY'
    WEATHER_CHOICES = (
        (CLOUDY, 'Cloudy'),
        (RAINY, 'Rainy'),
        (LIGHT_RAIN, 'Light rain'),
        (SUNNY, 'Sunny'),
        (SNOWY, 'Snowy'),
        (UNKNOWN, 'Unknown'),
    )

    SLIGHTLY_HEAVY = 'SLIGHTLY_HEAVY'
    HEAVY = 'HEAVY'
    GOOD = 'GOOD'
    BAD = 'BAD'
    TRACK_CONDITION_CHOICES = (
        (SLIGHTLY_HEAVY, 'Slightly heavy'),
        (HEAVY, 'Heavy'),
        (GOOD, 'Good'),
        (BAD, 'Bad'),
        (UNKNOWN, 'Unknown'),
    )

    STRAIGHT = 'STRAIGHT'
    LEFT = 'LEFT'
    RIGHT = 'RIGHT'
    OBSTACLE = 'OBSTACLE'
    COURSE_CHOICES = (
        (LEFT, 'Left'),
        (RIGHT, 'Right'),
        (STRAIGHT, 'Straight'),
        (OBSTACLE, 'Obstacle'),
        (UNKNOWN, 'Unknown'),
    )

    key = models.CharField(max_length=255, unique=True)
    racetrack = models.CharField(max_length=255, choices=RACETRACK_CHOICES, default=UNKNOWN)
    impost_category = models.CharField(max_length=255, choices=IMPOST_CATEGORY_CHOICES, default=UNKNOWN)
    surface = models.CharField(max_length=255, choices=SURFACE_CHOICES, default=UNKNOWN)
    course = models.CharField(max_length=255, choices=COURSE_CHOICES, default=UNKNOWN)
    distance = models.PositiveSmallIntegerField()
    number = models.PositiveSmallIntegerField()
    race_class = models.CharField(max_length=255, choices=RACE_CLASS_CHOICES, default=UNKNOWN)
    datetime = models.DateTimeField()
    weather = models.CharField(max_length=255, choices=WEATHER_CHOICES, default=UNKNOWN)
    track_condition = models.CharField(max_length=255, choices=TRACK_CONDITION_CHOICES, default=UNKNOWN)

    # 外枠か
    is_outside_racetrack = models.BooleanField()

    # (指定) [地]が出走できるレースで、かつ地方競馬所属の騎手が騎乗できる(特指)以外のレース
    is_regional_maiden_race = models.BooleanField()

    # (特指) JRAが認定した地方競馬の競走で第1着となった[地]が出走できるレースで、かつ地方競馬所属の騎手が騎乗できるレース
    is_winner_regional_horse_race = models.BooleanField()

    # [指定] 地方競馬所属の騎手が騎乗できるレース
    is_regional_jockey_race = models.BooleanField()

    # (混) 内国産馬にマル外が混合して出走できるレース
    is_foreign_horse_race = models.BooleanField()

    # (国際) 内国産馬に(外)および[外]が混合して出走できるレース
    is_foreign_trainer_horse_race = models.BooleanField()

    # (見習騎手) 若手騎手が騎乗できるレース
    is_apprentice_jockey_race = models.BooleanField()

    # (牝) 牝馬しか出走できないレース
    is_female_only_race = models.BooleanField()

    class Meta:
        db_table = 'races'
        get_latest_by = '-datetime'
        unique_together = ('racetrack', 'datetime')



from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='WebPage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('url', models.URLField()),
                ('html', models.TextField()),
                ('fingerprint', models.CharField(max_length=255, unique=True)),
                ('crawled_at', models.DateTimeField()),
            ],
            options={
                'db_table': 'webpages',
            },
        ),
    ]



# Generated by Django 2.1.5 on 2019-02-07 12:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('netkeiba', '0003_alter_race_remove_grade'),
    ]

    operations = [
        migrations.AlterField(
            model_name='race',
            name='race_class',
            field=models.CharField(choices=[('G1', 'G1'), ('G2', 'G2'), ('G3', 'G3'), ('OPEN', 'Open'), ('U1600', 'Under 1600'), ('U1000', 'Under 1000'), ('U500', 'Under 500'), ('MAIDEN', 'Maiden'), ('UNRACED_MAIDEN', 'Unraced Maiden'), ('UNKNOWN', 'Unknown')], default='UNKNOWN', max_length=255),
        ),
    ]



from django.db import models

from netkeiba.models.base import BaseModel


class Trainer(BaseModel):
    UNKNOWN = 'UNKNOWN'

    EAST = 'EAST'
    WEST = 'WEST'
    REGIONAL = 'REGIONAL'
    OVERSEAS = 'OVERSEAS'
    STABLE_CHOICES = (
        (EAST, 'East'),
        (WEST, 'West'),
        (REGIONAL, 'Regional'),
        (OVERSEAS, 'Overseas'),
        (UNKNOWN, 'Unknown')
    )

    key = models.CharField(max_length=255, unique=True)
    name = models.CharField(max_length=255)
    stable = models.CharField(max_length=255, choices=STABLE_CHOICES, default=UNKNOWN)

    class Meta:
        db_table = 'trainers'
